'-- HEADER  ----------------------------------------------------------------------------------------------------------------------
Rem'
	bbdoc:	Erweiterte Textformatierung
	about:	<b>Todo:</b><br>
					Weitere Formatierung wie Durchstreichen, Unterstreichen, Hintergrundfarbe ...<br><br>
					<b>Known Bugs:</b><br>
					&lt;AJ&gt; berechnet falsche Koordinaten, wenn es Leerzeilen im Text gibt (~n~n)<br><br>
					<b>Bentigte Module:</b><br>
					brl.pixmap<br>
					chaos.color<br>
					chaos.functions<br><br>
					<b>Empfohlenene Module:</b><br>
					bah.fontconfig (fr besseres Verhalten von BMax im Umgang mit Truetype Fonts)<br>
					chaos.bmpfont (fr besseres Verhalten von BMax im Umgang mit Bitmap Fonts)
End Rem
Module chaos.FormatText
ModuleInfo "Version: 1.00"
ModuleInfo "Modserver: BtbN"
ModuleInfo "OnAllOs: Yes"
ModuleInfo "Author: d-bug"
ModuleInfo "License: Public Domain"
ModuleInfo "Contact: d-bug@chaos-interactive.de"
ModuleInfo "Homepage: www.chaos-interactive.de"

ModuleInfo "History: 1.00 Erster Release"

Import chaos.functions
Import chaos.color
Import brl.pixmap


'-- INITIALISIERUNG  --------------------------------------------------------------------------------------------------------
SuperStrict

Private

	'Neuen Fontloader etablieren
	Type TChaosTextDriver Extends TFontLoader
		
		Global Alignment:Int
		Global CharData:TList = New TList
		
		Method LoadFont:TChaosTextFont (url:Object, size:Int, style:Int = SMOOTHFONT)
			If String(url) = "TChaosTextFont.Instance"
				If Not TChaosTextDriver.CharData.Count()
					Error("TChaosTextDriver.LoadFont","CharData is empty...")
					Return Null
				EndIf
				Return TChaosTextFont.Create (TChaosTextDriver.CharData, TChaosTextDriver.Alignment)
			EndIf
		End Method
	End Type
	AddFontLoader New TChaosTextDriver

	'----------------------------------------------------------------------------------------------------------
	'ChaosText Char Datentyp
	Type TChaosTextCharData
		
		Field Char:String
		Field Color:TColor
		Field Font:TImageFont
		Field Image:Object
		Field Line:Int
	
		Method New ()
			Char = ""
			Color = TColor.FromPaintColor ()
			Font = GetImageFont()
			Image = Null
		End Method
	End Type

	'----------------------------------------------------------------------------------------------------------
	'ChaosText Glyph
	Type TChaosTextGlyph Extends TGlyph

		Field _color:TColor
		Field _Char:String
		Field _Pixmap:TPixmap
		Field _Advance:Float
		Field _X:Int
		Field _Y:Int
		Field _Width:Float
		Field _Height:Int
		Field _Line:Int
		Field _OffsetX:Int
		Field _OffsetY:Int

		Method New ()
			_Pixmap = New TPixmap
			_Char = ""
			_Color = TColor.FromPaintColor ()
		End Method


		'Einen Glyph aus einem TChaosTextGlyphData-Object erstellen
		Function Create:TChaosTextGlyph (indata:Object,OffsetX:Int=0,OffsetY:Int=0, Nullglyph:TChaosTextGlyph)

			If indata = Null Return Null
			Local data:TChaosTextCharData = TChaosTextCharData(indata)
			If data.font.CharToGlyph(Asc(data.char)) = -1 Return NullGlyph
				
			Local glyph:TChaosTextGlyph = New TChaosTextGlyph
			glyph._color = data.color
			glyph._OffsetX = OffsetX
			glyph._OffsetY = OffsetY
			If data.image = Null
				glyph._char = data.char
				Local x:Int,y:Int,width:Int,height:Int
				Local dataglyph:TImageGlyph = data.font.LoadGlyph(data.font.CharToGlyph(Asc(glyph._char)))
				If dataglyph <> Null
					dataglyph.GetRect(x,y,width,height)
					glyph._X = x + OffsetX
					glyph._Y = y + OffsetY
					glyph._Width = width
					glyph._Height = height
					glyph._Advance = dataglyph.Advance()
					If glyph._char <> Chr(32)
						Local pixels:Object = dataglyph.Pixels()
						If TPixmap(pixels)
							glyph._Pixmap = TPixmap(Pixels).Copy()
						ElseIf TImage(Pixels)
							glyph._Pixmap = TImage(Pixels).Pixmaps[0].Copy()
						Else
							Error("TChaosTextGlyph.Create","Wrong Datatype for Pixels() glyph: <"+glyph._Char+">")
							Return NullGlyph
						EndIf
					EndIf
				EndIf
				Return glyph
			Else
				glyph._Char = ""
				If TPixmap(data.image)
					glyph._Pixmap = TPixmap(data.image).Copy()
				ElseIf TImage(data.image)
					glyph._Pixmap = TImage(data.image).Pixmaps[0].Copy()
				Else
					Error("TChaosTextGlyph.Create","Wrong Datatype for Pixels() glyph: <Image>")
					Return NullGlyph
				EndIf
				glyph._X = OffsetX
				glyph._Y = OffsetY
				glyph._Width = glyph._Pixmap.width
				glyph._Height = glyph._Pixmap.height
				glyph._Advance = glyph._Pixmap.width
				Return glyph
			EndIf
			Error("TChaosTextGlyph.Create","Unhandled Exception...")
			Return NullGlyph
		End Function

		'Pximap des Glyphs zurckgeben
		Method Pixels:TPixmap ()
			Return Self._Pixmap
		End Method
		
		'Breitenfaktor des Glyphs zurckgeben
		'Wird ermittelt aus Char-Padding und Char-Breite
		'In diesem Fall wird Advance vom Originalfont bernommen
		Method Advance:Float ()
			Return Self._Advance
		End Method
		
		'Abmae des Glyphs zurckgeben
		Method GetRect (x:Int Var, y:Int Var, width:Int Var, height:Int Var)
			x = Self._X
			y = Self._Y
			width = Int(Self._Width)
			height = Self._Height
		End Method
		
		
		'Farbe zurckgeben
		Method Color:TColor()
			Return Self._color
		End Method
		
	End Type

	'----------------------------------------------------------------------------------------------------------
	'ChaosText Font
	Type TChaosTextFont Extends TFont
		
		Field NullGlyph:TChaosTextGlyph
		
		Field Glyphs:TChaosTextGlyph[]
		Field LastGlyph:Int
		Field FontHeight:Int
		Field FontWidth:Int


		Method New()
			Nullglyph = New TChaosTextGlyph
		End Method


		'Pseudofont erstellen
		'Diese Fonts werden in der Reihenfolge der Chars des zughrigen Strings erstellt.
		'Normale Fonts werden in ASCII-Reihenfolge erstellt.
		Function Create:TChaosTextFont (datalist:TList, Alignment:Int)

			Local CurrentLine:Int, CurrentWidth:Int
			Local OffsetX:Int, OffsetY:Int

			If datalist <> Null
				If datalist.Count()
					Local font:TChaosTextFont = New TChaosTextFont
					For Local glyphdata:TChaosTextCharData = EachIn datalist
					
						'Linefeed addieren
						If glyphdata.Line > CurrentLine							
							'Vertikaler Linefeed
							Local diff:Int = glyphdata.Line - CurrentLine
							OffsetY:+(diff*font.FontHeight)
							TChaosText.CurrentHeight:+(diff*font.FontHeight)
							font.FontHeight = 0
							'Horizontaler LineFeed
							OffsetX:-CurrentWidth
							If CurrentWidth > font.FontWidth
								TChaosText.CurrentWidth = CurrentWidth
								font.FontWidth = CurrentWidth
							EndIf
							CurrentWidth = 0
							CurrentLine = glyphdata.Line
						EndIf
						
						Local index:Int = font.AddGlyph()
						font.Glyphs[index] = TChaosTextGlyph.Create(glyphdata, OffsetX, OffsetY, font.Nullglyph)
						If font.Glyphs[index]._Height > font.FontHeight
							font.FontHeight = font.Glyphs[index]._Height
						EndIf
						CurrentWidth:+font.Glyphs[index].Advance()
						font.Glyphs[index]._Line = CurrentLine
					Next
					TChaosText.CurrentHeight:+font.FontHeight
					If CurrentWidth > font.FontWidth
						TChaosText.CurrentWidth = CurrentWidth
					EndIf
					
					Select Alignment
						Case 1 ; font.AlignCenter ()
						Case 2 ; font.AlignRight ()
						Case 3 ; font.Alignjustify ()
					End Select
					
					Return font
				Else
					Error("TChaosTextFont.Load","Given datalist is empty...") 
					Return Null
				EndIf
			Else
					Error("TChaosTextFont.Load","Given datalist is Null")
					Return Null
			EndIf
			Error("TChaosTextFont.Load","Unhandled Exception...")
			Return Null
		End Function


		'Neuen Glyph an Liste anhngen
		Method AddGlyph:Int ()
			Self.Glyphs = Self.Glyphs[0..Self.Glyphs.Length+1]
			Return Self.Glyphs.Length-1
		End Method


		'Glyph Index ermitteln
		'Bei diesem FontLoader eigentlich egal, da alle Glyphs nach der Textreihenfolge
		'zurckgegeben werden mssen.
		Method CharToGlyph:Int (char:Int)
			Global syscolor:TColor
			If Not syscolor 
				syscolor = TColor.FromPaintColor ()
			Else
				syscolor.AsPaintColor ()
			EndIf
			Local Glyph:Int
			If Self.LastGlyph < Self.Glyphs.length
				Glyph = Self.LastGlyph
				Self.Glyphs[Glyph].Color().AsPaintColor ()
				Self.LastGlyph:+1
				If Self.LastGlyph = Self.Glyphs.length Self.LastGlyph = 0
				Return Glyph
			EndIf
			Return -1
		End Method


		'Glyph Instanz zurckgeben, wird von TImageFont bentigt
		'Hier wird auch gleichzeitig Farbe und Transparenz des Glyphs gesetzt.
		'Wir wollen ja schlielich lustige Effecte mit dem Text erzielen.
		Method LoadGlyph:TChaosTextGlyph (index:Int)
			If Self.Glyphs[index] <> Null Return Self.Glyphs[index]
			Return Self.NullGlyph
		End Method


		'Anzahl vorhandener Glyphs zurckgeben
		'Wieder eine von Marks sinnlosen Methoden :)
		Method CountGlyphs:Int ()
			Return Self.Glyphs.Length
		End Method


		'Anzahl vorhandener Zeilen zurckgeben
		Method CountLines:Int ()
			If Self.Glyphs
				Return Self.Glyphs[Self.Glyphs.length -1]._Line
			EndIf
		End Method
		
		
		'Zeile aus dem Text in ein Array bertragen
		Method GetLineArray:TChaosTextGlyph[] (line:Int)
			If line > Self.CountLines() Return Null
			Local result:TChaosTextGlyph[]
			For Local glyph:TChaosTextGlyph = EachIn Self.Glyphs
				If glyph._Line = line
					result = result[0..result.length+1]
					result[result.length-1] = glyph
				EndIf
			Next
			Return result
		End Method


		'Breite einer Zeile zurckgeben
		Method GetLineWidth:Int (linearray:TChaosTextGlyph[])
			Local result:Int = 0
			If linearray <> Null
				For Local glyph:TChaosTextGlyph = EachIn linearray
					If glyph = Self.Glyphs[Self.Glyphs.length-1]
						result:+Int(Self.Glyphs[Self.Glyphs.length-1]._width)
					Else
						result:+glyph.Advance()
					EndIf
				Next
			EndIf
			Return result
		End Method


		'Eine Liste mit allen Zeilen zurckgeben
		'Eine Zeile besteht hierbei aus einem TChaosTextGlyph Array
		Method GetLineList:TList ()
			If Self.CountLines() < 1 Return Null
			Local result:TList = New TList
			For Local line:Int = 0 To Self.CountLines()
				result.AddLast (Self.GetLineArray(line))
			Next
			Return result
		End Method
		
		
		'Breite des gesammten Textes zurckgeben
		Method GetTextWidth:Int ()
			Local lines:TList = Self.GetLineList()
			Local result:Int
			If lines <> Null
				For Local linearray:TChaosTextGlyph[] = EachIn lines		
					Local linewidth:Int = Self.GetLineWidth (linearray)
					If linewidth > result result = linewidth
				Next
			EndIf
			Return result
		End Method


		'Zentriert anordnen
		Method AlignCenter ()
			Local lines:TList = Self.GetLineList()
			Local twidth:Int = Self.GetTextWidth()
			If lines <> Null
				For Local linearray:TChaosTextGlyph[] = EachIn lines
					Local linewidth:Int = Self.GetLineWidth (linearray)
					If linewidth = twidth Continue
					Local diff:Int = (twidth/2) - (linewidth/2)
					For Local Glyph:TChaosTextGlyph = EachIn linearray
						Glyph._x:+diff
					Next
				Next
			EndIf
		End Method
		
		
		'Rechtsbndig anordnen
		Method AlignRight ()
			Local lines:TList = Self.GetLineList()
			Local twidth:Int = Self.GetTextWidth()
			If lines <> Null
				For Local linearray:TChaosTextGlyph[] = EachIn lines
					Local linewidth:Int = Self.GetLineWidth (linearray)
					If linewidth = twidth Continue
					Local diff:Int = twidth - linewidth
					For Local Glyph:TChaosTextGlyph = EachIn linearray
						Glyph._x:+diff
					Next
				Next
			EndIf
		End Method
		
	
		'Im Blocksatz anordnen
		Method AlignJustify ()

				'Leerzeichen zhlen
				Function CountSpaces:Int (linearray:TChaosTextGlyph[])
					Local result:Int
					For Local glyph:TChaosTextGlyph = EachIn linearray
						If glyph._char = Chr(32) result:+1
					Next
					Return result
				End Function


			Local lines:TList = Self.GetLineList()
			Local twidth:Float = Self.GetTextWidth()

			If lines <> Null
				
				'Spacebreiten neu berechnen
				For Local linearray:TChaosTextGlyph[] = EachIn lines
					If linearray = TChaosTextGlyph[](lines.Last()) Exit
					Local linewidth:Float = Self.GetLineWidth (linearray)
					If linewidth = twidth Continue
					Local spacediff:Float = (twidth - linewidth)	/ Float(CountSpaces(linearray))
					For Local glyph:TChaosTextGlyph = EachIn linearray
						If glyph._char = Chr(32)
							glyph._width:+spacediff
							glyph._Advance:+spacediff
						EndIf
					Next
				Next

				For Local linearray:TChaosTextGlyph[] = EachIn lines
					For Local glyph:TChaosTextGlyph = EachIn linearray
						If linearray <> TChaosTextGlyph[](lines.First())
							glyph._x:-glyph._OffsetX
							glyph._OffsetX = -(Self.GetLineWidth (TChaosTextGlyph[](lines.First()))*glyph._Line)
							glyph._x:+glyph._OffsetX
						EndIf
					Next
				Next

			EndIf
		End Method


		'Style des Fonts zurck geben
		'Wird hier vorerst nicht bentigt
		Method Style:Int ()
			Return 0
		End Method


		'Hhe des Fonts zurckgeben
		Method Height:Int ()
			Return Self.FontHeight
		End Method

	End Type




Type TChaosText

	Global List:TList = New TList
	Global CurrentWidth:Int
	Global CurrentHeight:Int

	Field _Text:String
	Field _Font:TImageFont
	Field _Width:Int
	Field _Height:Int

	Function ByObject:TChaosText (source:Object)
		For Local Search:TChaosText = EachIn TChaosText.List
			If String(source)
				If Search._Text = String(source) Return Search
			ElseIf TImageFont(source)
				If Search._Font = TImageFont(source) Return Search
			EndIf
		Next
		Return Null
	End Function

	Method Width:Int ()
		Return Self._width
	End Method

	Method Height:Int ()
		Return Self._height
	End Method

	'----------------------------------------------------------------------------------------------------
	'Text formatieren
	Function Format (Text:String Var, Font:TImageFont Var)
		Local FText:TChaosText = New TChaosText
		TChaosText.List.AddLast(FText)
		FText._Text = FText.Parse(Text)
		FText._Font = TImageFont.Load ("TChaosTextFont.Instance",0,SMOOTHFONT)
		FText._Width = TChaosText.CurrentWidth
		FText._Height = TChaosText.CurrentHeight
		Text = FText._Text
		Font = FText._Font 
	End Function

	'----------------------------------------------------------------------------------------------------

	Method Parse:String (Text:String)
	
			'Daten Zwischenspeicher
			Global CurrentColor:TColor = TColor.FromPaintColor ()
			Global CurrentFont:TImageFont = .GetImageFont()
			Global CurrentImage:Object = Null
			Global CurrentLine:Int

	
			'Funktionsinterner Tag Parser
			Function ParseTag:Int (text:String, pos:Int)
				If Text[pos..pos+1] <> "<" Return pos
				
				'Tag ausfiltern
				Local Tag:String = Text[pos..][0..(Text[pos..].Find(">")+1)].ToUpper()
				Local posOffset:Int

				'Zentriert
				If Tag = "<AC>"
					TChaosTextDriver.Alignment = 1
				EndIf

				'Rechtsbndig
				If Tag = "<AR>"
					TChaosTextDriver.Alignment = 2
				EndIf

				'Blocksatz
				If Tag = "<AJ>"
					TChaosTextDriver.Alignment = 3
				EndIf
				
				'Farbtag auswerten
				If Tag.Find("<COLOR") >= 0
					Local rgb:String[] = Split(Tag[Tag.Find("=")+1..(Tag.Find(">"))].Trim(),",")
					CurrentColor = TColor.FromValue (Int(rgb[0]),Int(rgb[1]),Int(rgb[2]))
				EndIf
				If Tag = "</COLOR>"
					CurrentColor = TColor.FromPaintColor ()
				EndIf
				
				'Alphatag auswerten
				Rem
				If Tag.Find("<ALPHA") >= 0
					CurrentColor.AdjustAlpha(Float(Tag[Tag.Find("=")+1..(Tag.Find(">"))].Trim()))
				EndIf
				If Tag = "</ALPHA>"
					CurrentColor.AdjustAlpha(GetAlpha())
				EndIf
				End Rem
				
				If Tag.Find ("<FONT") >= 0
					Local font:Object = HandleToObject(Int(Tag[Tag.Find("=")+1..(Tag.Find(">"))].Trim()))
					If TImageFont(font) CurrentFont = TImageFont(font)
				EndIf
				If Tag = "</FONT>"
					CurrentFont = .GetImageFont()
				EndIf
				
				If Tag.Find ("<IMG") >= 0
					Local image:Object = HandleToObject(Int(Tag[Tag.Find("=")+1..(Tag.Find(">"))].Trim()))
					If TImage(image)
						CurrentImage = TImage(image)
						posOffset = -1
					EndIf
				EndIf

				pos:+Tag.length+posOffset
				If Text[pos..pos+1] = "<"
					pos = ParseTag (Text,pos)
				EndIf
				Return pos
			End Function

		Local result:String

		'LineFeeds fromatieren
		Text.Replace("~n",Chr(10))
		
		For Local pos:Int = 0 To Text.length
			Local char:String = Text[pos..pos+1]
			If char = "<"
				pos = ParseTag (Text, pos)
				If pos+1 > Text.length Exit
				char = Text[pos..pos+1]
			EndIf
			If char = Chr(10)
				CurrentLine:+1
				Continue
			EndIf
			
			If Asc(char) >= 32 And Asc(char) < 256
				Local data:TChaosTextCharData = New TChaosTextCharData
				TChaosTextDriver.CharData.AddLast(data)
				data.Char = char
				data.Color = CurrentColor
				data.font = CurrentFont
				data.Image = CurrentImage
				data.Line = CurrentLine
				CurrentImage = Null
				result:+char
			EndIf
			
		Next
		Return result
	End Method

End Type
Public


'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Text formatieren
	returns:	Text ohne Tags als String und passenden Font
	about:		Diese Funktion wertet einen Text aus und fllt einen TImagefont mit dessen Daten.<br>
						Dies hat den Vorteil, dass nicht jedes Zeichen des Textes einzeln mit DrawText gezeichnet<br>
						werden muss, nur um ihn mit	verschiedenen Farben und Fonts zu versehen, was natrlich stark<br>
						auf die Performance gehen wrde.<br><br>
						Man kann nach dem Formatieren des Textes einfach mit SetImageFont(Rckgabefont) und<br>
						DrawText(Rckgabetext) den Text darstellen. Weiterhin kann so auch ohne groen	Aufwand<br>
						der formatierte Text mit SetScale, SetRotation, Sethandle	usw. dargestellt werden.<br><br>
						<b>Gltige Formatierungen:</b><br>
						<table>
							<tr><th>Formatierung</th><th>Beschreibung</th></tr>
							<tr><td>~n</td><td>Zeilenumbruch</td></tr>
							<tr><td>&lt;AR&gt;</td><td>Text wird rechtsbndig dargestellt</td></tr>
							<tr><td>&lt;AC&gt;</td><td>Text wird horizontal zentriert dargestellt</td></tr>
							<tr><td>&lt;AJ&gt;</td><td>Text wird im Blocksatz dargestellt</td></tr>
							<tr><td>&lt;ALPHA=a&gt;</td><td>Transparenz des folgenden Textes setzen (zwischen 0.0 und 1.0)</td></tr>
							<tr><td>&lt;/ALPHA&gt;</td><td>Transparenz zurcksetzen</td></tr>
							<tr><td>&lt;COLOR=rrr,ggg,bbb&gt;</td><td>Farbe des folgenden Textes setzen</td></tr>
							<tr><td>&lt;/COLOR&gt;</td><td>Farbe zurcksetzen</td></tr>
							<tr><td>&lt;FONT=handle&gt;</td><td>Font fr folgenden Text setzen (mit HandleFromObject ermittelt)</td></tr>
							<tr><td>&lt;/FONT&gt;</td><td>Font zurcksetzen</td></tr>
							<tr><td>&lt;IMG=handle&gt;</td><td>Image als Zeichen einfgen (mit HandleFromObject ermittelt)</td></tr>
						</table>
						Farbe und Transparenz werden mit der aktuellen Zeichenfarbe, bzw. Zeichentransparenz gemischt.<br><br>
						Natrlich hat das ganze auch Nachteile.<br>
						Da TextWidth() nicht in der Lage ist Zeichenbreiten in verschiedenen Zeilen vernnftig zu berechnen,<br>
						musste leider eine eigene Funktion her (siehe FormatTextWidth()).<br><br>
						Das gleiche gilt fr Mark Sibleys geniale Verarsche namens TextHeight() die nichts anderes macht als<br>
						die Hhe des aktuell gesetzten Fonts zurckzugeben. Diese hab ich auch durch eine neue Funktion<br>
						ersetzen mssen (siehe FormatTextHeight()).<br><br>
						Auerdem ist das ganze nicht unbedingt fr Realtime Textnderungen geeignet und war dafr auch nicht gedacht.
End Rem
Function FormatText (Text:String Var, Font:TImageFont Var)
	TChaosText.Format (Text, Font)
End Function


'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Breite des formatierten Textes ermitteln
	returns:	Breite als Integer
	about:		Also bergabeobject kann sowohl der mit FormatText() zurckgegebne Text, als auch
						der dazu gehrende Font angegeben werden.
End Rem
Function FormatTextWidth:Int (Source:Object)
	Local FText:TChaosText = TChaosText.ByObject (Source)
	If FText = Null Return 0
	Return FText.Width()
End Function


'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Hhe des formatierten Textes ermitteln
	returns:	Hhe als Integer
	about:		Also bergabeobject kann sowohl der mit FormatText() zurckgegebne Text, als auch
						der dazu gehrende Font angegeben werden.
End Rem
Function FormatTextHeight:Int (Source:Object)
	Local FText:TChaosText = TChaosText.ByObject (Source)
	If FText = Null Return 0
	Return FText.Height()
End Function
