'-- HEADER  ----------------------------------------------------------------------------------------------------------------------
Rem'
	bbdoc:	Kleines Modul mit oft benutzten Standard-Funktionen
EndRem
Module chaos.functions
ModuleInfo "Version: 1.00"
ModuleInfo "Modserver: BtbN"
ModuleInfo "OnAllOs: Yes"
ModuleInfo "Author: d-bug"
ModuleInfo "License: Public Domain"
ModuleInfo "Contact: d-bug@chaos-interactive.de"
ModuleInfo "Homepage: www.chaos-interactive.de"

ModuleInfo "History: 1.01 ScreenShot () hinzu"
ModuleInfo "History: 1.00 Erster Release"

'-- INITIALISIERUNG  --------------------------------------------------------------------------------------------------------
SuperStrict

Import brl.blitz
Import brl.ramstream
Import brl.system
Import brl.max2d
Import brl.linkedlist
Import brl.pixmap
Import brl.pngloader

Const CIERROR_NONOTIFY:Int	= %1001
Const CIERROR_CRITICAL:Int	= %0110

'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Einen Screenshot erstellen
	returns:	---
	about:		Erstellt einen Screenshot und speichert ihn unter dem angegebenen Pfad ab
EndRem
Function Screenshot (path:String)
	SavePixmapPNG (GrabPixmap(0,0,GraphicsWidth(),GraphicsHeight()),path,5)
End Function

'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Fehler-Ausgabe in der Konsole
	returns:	---
	about:		Diese Funktion gibt einen Text im Deebugmodus in der Konsole aus. Sie dient nur
						der Vereinfachung der Fehler-Ausgabe
EndRem

Function Error (func:String, message:String, flags:Int = 0)
	?Debug
		WriteStdout "'"+ AppTitle + "' @"+MilliSecs()+" :: "+func+" > "+message+"~n"
		If flags & CIERROR_CRITICAL
			WriteStdout "Critical Error... '"+ AppTitle +"' terminated..."
			End
		EndIf
		Return
	?
	If Not flags & CIERROR_NONOTIFY
		Notify ("'"+ AppTitle +"' @"+MilliSecs()+"~n"+func+"~n"+message)
		If flags & CIERROR_CRITICAL End
	EndIf
End Function

'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Pfadprfung fr Dateien.
	returns:	Einen Pfad zur datei oder "failed" wenn die Datei nicht gefunden wurde
	about:		Diese Funktion prft eine Pfadangabe auf das Vorhandensein der Datei.
						Wird die Datei nicht gefunden gibt sie "failed" zurck, andernfalls den
						geprften Pfad.
EndRem
Function ValidatePath:String (url:String)
	url = url.ToLower()
	If url.Find("incbin::") > -1 Return url
	FixPath(url)
	If FileType(url) Then Return url
	Return "failed"
End Function

'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		RGB Wert prfen.
EndRem
Function ValidateRGB:Int (value:Int)
	If value < 0 value = 0
	If value > 255 value = 255
	Return value
End Function
	
'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Alpha Wert prfen.
EndRem
Function ValidateA:Float (value:Float)
	If value < 0 value = 0
	If value > 1.0 value = 1.0
	Return value
End Function


'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Einen String aufteilen.
	returns:	String Array
	about:		Diese Funktion teilt einen gegebenen String auf und fllt mit den
						Teilstcken ein String-Array. Als Teiler kann eine beliebige Zeichenfolge
						angegeben werden. Standardmig wird ein Leerzeichen als Teiler definiert.
EndRem
Function Split:String[] (source:String,separator:String=Chr(32))
	Local position:Int,oldposition:Int,output:String[]
	Repeat
		position = source.find(separator,position)+1
		output = output[..output.length+1]
		output[output.length-1] = source[oldposition..position-1]
		If position = 0 output[output.length-1] = source[oldposition..]
		oldposition = position+separator.length-1
	Until position = 0
	Return output
End Function


'----------------------------------------------------------------------------------------------------
Rem'
	bbdoc:		Lineare Interpolation.
EndRem
Function LInterpolate:Float (value:Float, valmax:Float, retmax:Float)
	Return (value * (retmax / valmax))
End Function




